/*
** toybox.c: container exploration toy
**
** This program implements a primitive "container"
** under the Linux 4.4 kernel. It requires an
** existing target directory (which will become
** the container's root directory), populated with
** (at a minimum) the bin directory and a bin/sh
** binary. toybox will create any other directories
** that it requires for it's operation.
**
** Usage: toybox [-h hostname] [-H] rootdir
** 
** Where: -h hostname sets the container hostname to specified value
**	  -H r/o bind mounts host /bin, /sbin, /lib, /lib64 and /usr,
**   and  rootdir is the name of the directory which toybox will use
**	    as it's own root directory
**
** toybox directory structure:
**
**   /
**   :
**   +- /bin		required, prepopulated, must include /bin/sh
**   :				  (may be r/o bind mounted with host /bin)
**   :
**   +- /dev		constructed, toybox will bind mount host /dev to this mountpoint
**   :
**   +- /etc		optional, may hold prepopulated config files
**   :
**   +- /lib		optional, may be r/o bind mounted with host /lib
**   :
**   +- /lib64		optional, may be r/o bind mounted with host /lib64
**   :
**   +- /mnt		constructed, toybox will hold temp mounts here
**   :  :
**   :  +- /mnt/.proc	constructed, used by toybox to temporarily mount host /proc
**   :  :
**   :  +- /mnt/.root	constructed, used by toybox to temporarily mount host /
**   :  :
**   :  +- /mnt/.sys	constructed, used by toybox to temporarily mount host /sys
**   :
**   +- /proc		constructed, toybox will mount procfs to this mountpoint
**   :
**   +- /sbin		optional, may hold prepopulated binaries
**   :				  (may be r/o bind mounted with host /sbin)
**   :
**   +- /sys		constructed, toybox will mount sysfs to this mountpoint
**   :
**   +- /tmp		constructed, toybox will mount tmpfs to this mountpoint
**   :
**   +- /usr		optional, may hold prepopulated directories, binaries
**   :				  (may be r/o bind mounted with host /usr)
**   :
**   +- /var		optional, may hold prepopulated data
**
**
** NB: compile with -D_GNU_SOURCE
**
** $Id: toybox.c,v 1.34 2023/08/24 22:57:29 lpitcher Exp lpitcher $
**
** Copyright (C) 2023  Lew Pitcher
**
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2 of the License, or
** (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License along
** with this program; if not, write to the Free Software Foundation, Inc.,
** 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
**
** Lew Pitcher <lew.pitcher@digitalfreehold.ca>
**
*/
 
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sched.h>
#include <string.h>
#include <fcntl.h>
#include <errno.h>
#include <stdarg.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/mount.h>
#include <sys/wait.h>

/* pivot_root() function prototype not supplied by standard headers */
extern int pivot_root(const char *new_root, const char *put_old);

#define	CONTAINER	(CLONE_NEWUSER	|	\
			 CLONE_NEWNS	|	\
			 CLONE_NEWPID	|	\
			 CLONE_NEWUTS	|	\
			 CLONE_NEWNET	|	\
			 CLONE_NEWIPC)
/*
** NB:	CLONE_NEWUSER	gives us a new user namespace, and the process
**			gains CAP_SYS_ADMIN
**	CLONE_NEWNS	gives us a new mount namespace, which requires
**			the process have CAP_SYS_ADMIN to manipulate
**	CLONE_NEWPID	gives us a new PID namespace, which
**			gives access to the /proc filesystem,
**			and the special properties of PID 1
**			this namespace also requires the manipulating
**			process to have CAP_SYS_ADMIN
**	CLONE_NEWUTS	gives us a new UTS namespace, which requires
**			the process have CAP_SYS_ADMIN to manipulate
**	CLONE_NEWNET	gives us a new NETWORK namespace, which
**			gives us access to an isolated set of network
**			resources. Requires CAP_SYS_ADMIN
**	CLONE_NEWIPC	gives us a new IPC namespace, which
**			gives us access to an isolated set of Sys V IPC
**			and POSIX message queue resources. Requires
**			CAP_SYS_ADMIN
**	CLONE_NEWCGROUP	is not implemented in kernel 4.4
*/

#define	FLAG_HOSTENV	(1<<0)

static int	toybox(const char *RootDir, const char *HostName, unsigned int toyflags);
static int 	assumeRoot(uid_t oldUid, gid_t oldGid);
static int 	isolateFs(const char *RootDir, unsigned int toyflags);
static int	idleProcess(const char *HostName);
static int	initProcess(const char *HostName);
static int	recbind_ro(const char *source, const char *target);
static int 	sysMsg(const char *sev, int errnum, const char *msg, ...);

#define		ERROR(...) 	(sysMsg("Error",errno,__VA_ARGS__),0)
#define		WARNING(...)	sysMsg("Warning",errno,__VA_ARGS__)

int main(int argc, char *argv[])
{
  int	success = 1,		/* assume success */
	opt;
  char	*HostName = NULL;
  unsigned int toyflags = 0;

  while ((opt = getopt(argc,argv,"h:H")) != -1)
  {
    switch (opt)
    {
      case 'h':
	free(HostName);
	HostName = strdup(optarg);
	break;

      case 'H':
	toyflags |= FLAG_HOSTENV;
	break;


      default:
	success = 0;
	break;
    }
  }

  if ((success) && (argc - optind == 1))
    success = toybox(argv[optind],HostName,toyflags);
  else
    fprintf(stderr,"Usage: %s [-h hostname] [-H] rootdir\n",argv[0]);

  return success?EXIT_SUCCESS:EXIT_FAILURE;
}

/*
** toybox(const char *RootDir, const char *HostName, unsigned int toyflags)
**
** Create and execute our toy container by unshare()ing
** its namespaces, changing its uid and gid, building its
** root filesystem and executing its idle and init processes.
**
** Accepts:	const char * pointer to root directory path string
**		const char * pointer to hostname string
**		an unsigned int containing various execution flags
**		  FLAG_HOSTENV will indicate whether (1) or not (0)
**		  to mount additional directories from the host
**		  filesystem
**
** Returns:	int 1 if successful in creating container
**		int 0 if container creation failed
*/
static int toybox(const char *RootDir, const char *HostName, unsigned int toyflags)
{
  int	success = 1;	/* assume success */
  uid_t	host_uid;
  gid_t	host_gid;

  umask(0);
  host_uid = getuid();	/* must do before unshare() as uid changes within unshare */
  host_gid = getgid();	/* must do before unshare() as gid changes within unshare */

  /* change our namespaces */
  if (unshare(CONTAINER))
    success = ERROR("unshare()");

  /* change our uid, gid to 0,0 */
  if (success)
    success = assumeRoot(host_uid,host_gid);

  /* set up our isolated root directory */
  if (success)
    success = isolateFs(RootDir,toyflags);

  /* run our idle and init processes within our isolated root filesystem*/
  if (success)
    success = idleProcess(HostName);

  return success;
}

/*
** assumeRoot(oldUid, oldGid)
**
** Update the process entry in /proc to map our 
** outside-the-box uid and gid to the inside-the-box
** root user uid and gid, matching the effects
** of our new CLONE_NEWUSER capabilities.
**
** Accepts:	uid_t oldUid - our outside-the-box UID,
**		gid_t oldGid - our outside-the-box GID
**
** Returns: 	int 1  - success
**	    	int 0  - failure
*/
static int assumeRoot(uid_t oldUid, gid_t oldGid)
{
  int	success = 1,	/* assume success */ 
	fd;
  FILE *map;

  /*
  ** Map our uid into the container's root uid (UID 0)
  */
  if (map = fopen("/proc/self/uid_map","w"))
  {
    fprintf(map,"0 %lu 1\n",(unsigned long)oldUid);
    fclose(map);
  }
  else success = ERROR("fopen(\"/proc/self/uid_map\")");

  /*
  ** Some kernels require /proc/self/setgroups set to "deny"
  ** before we can remap our GID
  */
  if (success  && ((fd = open("/proc/self/setgroups",O_WRONLY)) >= 0))
  {
    write(fd,"deny",4);
    close(fd);
  }

  if (success)
  {
    /*
    ** Map our gid into the container's root gid (GID 0)
    */
    if (map = fopen("/proc/self/gid_map","w"))
    {
      fprintf(map,"0 %lu 1\n",(unsigned long)oldGid);
      fclose(map);
    }
    else success = ERROR("fopen(\"/proc/self/uid_map\")");
  }

  return success;
}

/*
** isolateFs(const char *RootDir, unsigned int toyflags)
**
** Build our container's root filesystem by creating
** required subdirectories in target rootdir, then
** mount()ing various requisite resources on those
** subdirectories.
**
** Accepts:	a const char pointer to the pathname of the root filesystem
**		an unsigned int containing various execution flags
**		  FLAG_HOSTENV will indicate whether (1) or not (0)
**		  to mount additional directories from the host
**		  filesystem
**
** Returns:	int 1 if successful creation of container
**		 root filesystem hierarchy
**		int 0 if container root filesystem cannot
**		 be properly built
*/
static int isolateFs(const char *RootDir, unsigned int toyflags)
{
  int	success = 1;	/* assume success */

  /* Ensure that none of our actions here affect host mountspace */
  if (mount("none","/",NULL,MS_REC|MS_PRIVATE,NULL))
    success = ERROR("mount(,\"/\",,)");

  /* make our root directory into a mountpoint in prep for pivot_root */
  if (success && mount(RootDir,RootDir,NULL,MS_BIND|MS_NOSUID,NULL))
    success = ERROR("mount(\"%s\",\"%s\",NULL,MS_BIND|MS_NOSUID,NULL)",RootDir,RootDir);

  /* Continue operations from within our intended root directory */
  if (success && chdir(RootDir))
    success = ERROR("chdir(\"%s\")",RootDir);

  /* Remove and recreate any of the unpopulated, required subdirectories */
  if (success)
  {
    if (unlink("dev") && rmdir("dev"))			/* remove any residual /dev directory */
      WARNING("rmdir(\"dev\")");

    if (unlink("mnt/.proc") && rmdir("mnt/.proc"))	/* remove any residual /mnt/.proc directory */
      WARNING("rmdir(\"mnt/.proc\")");

    if (unlink("mnt/.root") && rmdir("mnt/.root"))	/* remove any residual /mnt/.root directory */
      WARNING("rmdir(\"mnt/.root\")");

    if (unlink("mnt/.sys") && rmdir("mnt/.sys"))	/* remove any residual /mnt/.sys directory */
      WARNING("rmdir(\"mnt/.sys\")");

    if (unlink("mnt") && rmdir("mnt"))			/* remove any residual /mnt directory */
      WARNING("rmdir(\"mnt\")");

    if (unlink("proc") && rmdir("proc"))		/* remove any residual /proc directory */
      WARNING("rmdir(\"proc\")");

    if (unlink("sys") && rmdir("sys"))			/* remove any residual /sys directory */
      WARNING("rmdir(\"sys\")");

    if (unlink("tmp") && rmdir("tmp"))			/* remove any residual /tmp directory */
      WARNING("rmdir(\"tmp\")");
  }

  if (success && mkdir("dev",0755))			/* create a new, empty /dev directory */
    success = ERROR("mkdir(\"dev\")");			/* to install our devices in */

  if (success && mkdir("mnt",0755))			/* create a new, empty /mnt directory */
    success = ERROR("mkdir(\"mnt\")");			/* to host some temporary mounts */

  if (success && mkdir("mnt/.proc",0755))		/* create a new, empty /mnt/.proc directory */
    success = ERROR("mkdir(\"mnt/.proc\")");		/* to temporarily mount a procfs to 	    */

  if (success && mkdir("mnt/.root",0751))		/* create a new, empty /mnt/.root directory */
    success = ERROR("mkdir(\"mnt/.root\")");		/* to temporarily mount our old root dir to */

  if (success && mkdir("mnt/.sys",0751))		/* create a new, empty /mnt/.sys directory */
    success = ERROR("mkdir(\"mnt/.sys\")");		/* to temporarily mount our old /sys dir to */

  if (success && mkdir("proc",0755))			/* create a new, empty /proc directory */
    success = ERROR("mkdir(\"proc\")");			/* to mount our procfs to */

  if (success && mkdir("sys",0755))			/* create a new, empty /sys directory */
    success = ERROR("mkdir(\"sys\")");			/* to mount our sysfs to */

  if (success && mkdir("tmp",01777))			/* create a new, empty /tmp directory */
    success = ERROR("mkdir(\"tmp\")");			/* to mount a tmpfs to */


  /* Populate our /dev directory with direct link to host /dev */
  if (success && mount("/dev","dev",NULL,MS_BIND|MS_REC,NULL))
    success = ERROR("mount(\"dev\")");

  if (toyflags & FLAG_HOSTENV)
  {
    if (success) success = recbind_ro("/bin","bin");
    if (success) success = recbind_ro("/lib","lib");
#ifdef LIB64
    if (success) success = recbind_ro("/lib64","lib64");
#endif
    if (success) success = recbind_ro("/sbin","sbin");
    if (success) success = recbind_ro("/usr","usr");
  }

  /* Populate our (temporary) proc with direct link to host /proc */
  if (success && mount("/proc","mnt/.proc",NULL,MS_BIND|MS_REC,NULL))
    success = ERROR("mount(\"mnt/.proc\")");

  /* Populate our /sys directory with direct link to host /sys */
  if (success && mount("/sys","mnt/.sys",NULL,MS_BIND|MS_REC,NULL))
    success = ERROR("mount(\"mnt/.sys\")");

  /* Give ourselves a tempfs for our /tmp directory */
  if (success && mount("none","tmp","tmpfs",
			MS_NOSUID|MS_NODEV|MS_NOEXEC|MS_NOATIME,
			"size=16m,nr_inodes=4k,mode=770"))
    success = ERROR("mount(\"tmp\")");

  /* Now, Isolate our directory space and occupy the toy box */
  if (success && pivot_root(".","mnt/.root"))
    success = ERROR("pivot_root(\".\",\"mnt/.root\")");

  /* pivot_root(2) recommends that we chdir("/") to ensure correct cwd */
  if (success && chdir("/"))
    success = ERROR("chdir(\"/\")");

  /* Discard our tie to the host root directory */
  if (success && umount2("/mnt/.root",MNT_DETACH))
    success = ERROR("umount2(\"/mnt/.root\")");

  /* Discard the temporary directory on which we saved host root */
  if (success && unlink("/mnt/.root") && rmdir("/mnt/.root"))
    WARNING("rmdir(\"/mnt/.root\")");

  return success;
}

/*
** idleProcess(toyParms_t toyParms)
**
** Execute container's "idle" (PID 0) 
** and "init" (PID 1) processes
**
** Accepts:	a const char pointer to new hostname, or NULL
**
** Returns:	int 1 if successful execution of container
**		int 0 if failure of init process
*/
static int idleProcess(const char *HostName)
{
  int	success = 1,		/* assume success */
	child_status;
  pid_t	child;

  switch (child = fork())
  {
    case -1:	/* still in parent: fork() failed */
      success = ERROR("fork()");
      break;

    case 0:	/* in child, which takes the init process role */
      success = initProcess(HostName);
      break;

    default:	/* in parent, which takes on the idle process role */
      wait(&child_status);
      if (!WIFEXITED(child_status) || (WEXITSTATUS(child_status) != 0))
	success = 0;

      /* unmount our private copies of mounted filesystems */
      umount2("/tmp",MNT_DETACH);	/* not strictly necessary */
      umount2("/dev",MNT_DETACH);	/* not strictly necessary */
      umount2("/sys",MNT_DETACH);	/* not strictly necessary */
      umount2("/proc",MNT_DETACH);	/* not strictly necessary */
  }

  return success;
}

/*
** initProcess(const char *HostName)
**
** Execute the container's init (PID 1) process
**
** Accepts:	a const char pointer to new hostname, or NULL
**
** Returns:	int 1 if successful execution of container
**		int 0 if failure of init process
**
** Note:	This function will generate warning messages
**		 under certain circumstances. These informational
**		 messages can be ignored.
*/
static int initProcess(const char *HostName)
{
  int	success = 1;		/* assume success */
  char	*envp[] = {"PATH=/sbin:/bin:/usr/sbin:/usr/bin:",NULL},
	*argv[] = {"-sh",NULL};

  /*
  ** >>>>>>>>>> Mount /proc and discard /mnt/.proc <<<<<<<<<<
  */
  /* Populate our /proc directory */
  if (mount("none","/proc","proc",MS_REC,NULL))
    success = ERROR("mount(\"/proc\")");

  /* Discard our tie to the host /proc directory */
  if (success && umount2("/mnt/.proc",MNT_DETACH))
    success = ERROR("umount2(\"/mnt/.proc\")");

  /* Discard the temporary directory on which we saved host proc */
  if (success && unlink("/mnt/.proc") && rmdir("/mnt/.proc"))
    WARNING("rmdir(\"mnt/.proc\")");


  /*
  ** >>>>>>>>>> Mount /sys and discard /mnt/.sys <<<<<<<<<<
  */
  /* Mount a true sysfs to our (isolated) /sys directory */
  if (success && mount("none","/sys","sysfs",0,NULL))
    success = ERROR("mount(\"/sys\")");

  /* Discard our tie to the host /sys directory */
  if (success && umount2("/mnt/.sys",MNT_DETACH))
    success = ERROR("umount2(\"/mnt/.sys\")");

  /* Discard the temporary directory on which we saved host /sys */
  if (success && unlink("/mnt/.sys") && rmdir("/mnt/.sys"))
    WARNING("rmdir(\"/mnt/.sys\")");


  /*
  ** >>>>>>>>>> Set hostname, run init process (shell) <<<<<<<<<<
  */
  /* If required, set our container's UTS hostname */
  if (success && HostName)
    if (sethostname(HostName,strlen(HostName)))
      WARNING("sethostname(\"%s\",%lu)",HostName,(unsigned long)strlen(HostName));

  /* Now, run a shell as our init process */
  if (success && execvpe("/bin/sh",argv,envp))
       success = ERROR("execle(\"/bin/sh\")");

  return success;
}

/*
** recbind_ro(const char *source, const char *target)
**
** Read-only BIND mount a given source directory onto the given target directory
**
** Accepts:	a const char pointer to the path to the source directory,
**		a const char pointer to the path to the target directory
**
** Returns:	int 1 if successful BIND mount
**		int 0 if unsuccessful BIND mount
**
** Note:	This function will generate a warning message
**		if the BIND mount cannot be remounted as read-only.
*/
static int recbind_ro(const char *source, const char *target)
{
  int success = 1;

  unlink(target);		/* get rid of regular file with target name */
  mkdir(target,0);		/* create directory with target name, no permissions */

  if (mount(source,target,NULL,MS_BIND|MS_REC,NULL))
    success = ERROR("mount(\"%s\",\"%s\",NULL,MSBIND_MS_REC,NULL)",source,target);
  else
    if (mount(source,target,NULL,MS_REMOUNT|MS_BIND|MS_REC|MS_RDONLY,NULL))
      WARNING("mount(\"%s\",\"%s\",MS_REMOUNT|MS_BIND|MS_REC|MS_RDONLY,NULL)",source,target);

  return success;
}

/*
** sysMsg(const char *sev, int errnum, const char *msg, ...)
**
** Report a syscall error
**
** Accepts:	a const char pointer to a severity string,
**		an integer, derived from the errno returned by the syscall,
**		a const char pointer to a message format string, and
**		optional arguments as required to satisfy the message format string
**
** Returns:	int 0
**
** Note: returns a discardable int, so that function can be used in a comma operator list
*/
static int sysMsg(const char *sev, int errnum, const char *msg, ...)
{
  va_list ap, pp;
  char	dummy,
	*buf;
  int	buflen;

  va_start(ap,msg);

  va_copy(pp,ap);
  buflen = 2 + vsnprintf(&dummy,sizeof dummy, msg, pp);

  if (buf = malloc(buflen))
  {
    va_copy(pp,ap);
    vsnprintf(buf,buflen,msg,pp);
  }

  va_end(ap);

  fprintf(stderr,"%s: %s returned \"%s\"\n",sev,buf?buf:"a system call",strerror(errnum));
  free(buf);

  return 0;
}

/*
** $Log: toybox.c,v $
** Revision 1.34  2023/08/24 22:57:29  lpitcher
** redesigned error, warning reporting
** changed all uses of SYSERROR to new error, warning reporting
**
** Revision 1.33  2023/08/24 20:02:01  lpitcher
** corrected "r/o bind mount" logic
**
** Revision 1.32  2023/08/21 19:48:42  lpitcher
** generalize execution flags
**
** Revision 1.31  2023/08/21 17:51:31  lpitcher
** conditional compilation around /lib64 bind mount
** #define LIB64 to enable
**
** Revision 1.30  2023/08/21 17:48:56  lpitcher
** more comment cleanup
**
** Revision 1.29  2023/08/21 17:45:38  lpitcher
** comment cleanup
**
** Revision 1.28  2023/08/21 17:40:10  lpitcher
** additional comments re r/o bind mounts
**
** Revision 1.27  2023/08/21 17:34:00  lpitcher
** added -H flag to demand bind mount of additional host dirs
** on demand, r/o bind mount host /bin, /sbin, /lib, /lib64
** and /usr to corresponding toybox mountpoints.
**
** Revision 1.26  2023/08/11 15:07:44  lpitcher
** root shell invoked as "login" shell
**
** Revision 1.25  2023/08/10 15:13:43  lpitcher
** removed tParameters. simplified all parameters to string arguments
** simplified final process argument success test, arg count test
** use stdio to update proc uid_map and gid_map
**
** Revision 1.24  2023/07/11 01:06:15  lpitcher
** minor comment cleanup
**
** Revision 1.23  2023/05/05 18:27:20  lpitcher
** added GPL v2 licence
**
** Revision 1.22  2023/01/27 01:59:47  lpitcher
** reverted code that creates own dev entries.
** will continue with valid code until we determine
** what caused the toydev failures
** This version logically identical to 1.20
**
** Revision 1.20  2023/01/22 14:30:34  lpitcher
** pass name of root directory in toyopts_t toyOptions structure
** remove rootdir reference from function declarations and definitions
** in recognition of expanded function of tOptions control block,
**   renamed tOptions to tParameters
**   renamed toyOpts_t to toyParms_t
**   renamed toyOpts to toyParms
**   renamed toyOptions to toyParameters
**   renamed FLAG_HOSTNAME to PARM_HOSTNAME
**
** Revision 1.19  2023/01/20 21:38:15  lpitcher
** after pivot_root(), chdir("/") to ensure correct cwd
** at idle termination, unmount our private copies of mounted filesystems
** pass toyopts_t toyParameters by value to subordinate functions
**
** Revision 1.18  2023/01/16 00:03:48  lpitcher
** added more commentary around directory manipulation
** isolate and mount our own sysfs to /sys
**
** Revision 1.17  2023/01/09 16:05:08  lpitcher
** use absolute paths once we pivot_root
**
** Revision 1.16  2023/01/09 15:44:40  lpitcher
** final cleanup of unused vars - fixed some unintended compile errors
**
** Revision 1.15  2023/01/09 15:43:44  lpitcher
** clean up unused vars
**
** Revision 1.14  2023/01/09 15:42:20  lpitcher
** simplified all isolateFs() tests
**
** Revision 1.13  2023/01/09 15:35:43  lpitcher
** simplified all initProcess() tests
**
** Revision 1.12  2023/01/09 15:31:37  lpitcher
** simplified sethostname call
**
** Revision 1.11  2023/01/09 15:29:30  lpitcher
** adjusted init PATH
**
** Revision 1.10  2023/01/09 13:31:46  lpitcher
** expanded container creation commentary
**
** Revision 1.9  2023/01/09 13:21:16  lpitcher
** replaced flags and hostname with toyParms_t structure
** set flag in toyParms_t for hostname when given
** use flag in toyParms_t to determine hostname when needed
**
** Revision 1.8  2023/01/09 07:29:03  lpitcher
** allow empty hostname to reset UTS hostname
**
** Revision 1.7  2023/01/09 07:00:21  lpitcher
** correct spelling errors
**
** Revision 1.6  2023/01/09 06:42:23  lpitcher
** propogate flags variable to assumeRoot(), isolateFs(), idleProcess(), and initProcess()
** global rename status variable to success, to better reflect its purpose
**
** Revision 1.5  2023/01/09 06:02:47  lpitcher
** refactored toybox() logic
** added parens to ERROR macro
**
** Revision 1.4  2023/01/09 05:43:19  lpitcher
** added empty hostname check to initProcess
** added descriptive comments for each function
**
** Revision 1.3  2023/01/09 05:30:53  lpitcher
** added sethostname() to initProcess()
** added hostname parameter to idleProcess(), initProcess()
** pass hostname to idleProcess, initProcess
**
** Revision 1.2  2023/01/09 05:18:46  lpitcher
** added RCS log
**
*/
